# frozen_string_literal: true
require "spec_helpers"

describe Wayfarer::Job do
  subject(:job) { Class.new(Wayfarer::Job) }

  describe "::prepare" do
    it "copies the router" do
      job.class_eval do
        route.host "example.com", to: :foobar
      end

      prepared_job = job.prepare

      expect(prepared_job.router).not_to be job.router

      uri = URI("http://example.com")
      expect(prepared_job.router.routes?(uri)).to be true
    end

    it "copies locals" do
      job.class_eval do
        let(:foo) { { bar: :qux } }
        let(:baz) { 42 }
      end

      prepared_job = job.prepare

      expect(prepared_job.locals).not_to be job.locals

      expect(prepared_job.locals[:foo]).to be_a Concurrent::Hash
      expect(prepared_job.locals[:baz]).to be_a Concurrent::AtomicFixnum
    end

    it "defines instance accessors for locals" do
      job.class_eval do
        let(:foo) { { bar: :qux } }
        let(:baz) { 42 }
      end

      prepared_job = job.prepare

      expect(prepared_job).to respond_to(:foo)
      expect(prepared_job).to respond_to(:baz)
    end
  end

  describe "::config" do
    it "allows manipulating the configuration" do
      expect {
        job.config { |c| c.http_adapter = :selenium }
      }.to change { job.config.http_adapter }.to(:selenium)
    end

    it "does not manipulate the global configuration" do
      expect {
        job.config { |c| c.http_adapter = :selenium }
      }.not_to change { Wayfarer.config.http_adapter }
    end
  end

  describe "#halt" do
    it "sets the halting flag" do
      job_instance = job.new

      expect {
        job_instance.send(:halt)
      }.to change { job_instance.halts? }.from(false).to(true)
    end
  end

  describe "#stage" do
    it "stages URIs" do
      job_instance = job.new
      job_instance.page = Page.new(uri: URI("https://yahoo.com/qux"))

      uris = %w(
        http://google.com
        http://example.com
      )

      expected = uris.map { |u| URI(u) }

      expect {
        job_instance.send(:stage, *uris)
      }.to change { job_instance.staged_uris }.from([]).to(expected)
    end

    it "expands relative URIs" do
      job_instance = job.new
      job_instance.page = Page.new(uri: URI("https://yahoo.com/qux"))

      uris = %w(
        /foo/bar
        bar/qux.html
        barfoo
      )

      expected = %w(
        https://yahoo.com/qux/foo/bar
        https://yahoo.com/qux/bar/qux.html
        https://yahoo.com/qux/barfoo
      ).map { |u| URI(u) }

      expect {
        job_instance.send(:stage, *uris)
      }.to change { job_instance.staged_uris }.from([]).to(expected)
    end

    context "with non-HTTP(s) URIs" do
      it "ignores them and does not raise an exception" do
        job_instance = job.new

        expect {
          job_instance.send(:stage, "ftp://example.com")
        }.not_to raise_error

        expect(job_instance.staged_uris).to be_empty
      end
    end
  end
end
