# frozen_string_literal: true
require "spec_helpers"

describe Wayfarer::Routing::Router do
  subject(:router) { Router.new }

  describe "#forbid" do
    it "forbids URIs" do
      router.forbid(host: "example.com")
      expect(router.forbids?(URI("http://example.com"))).to be true

      router.forbid { host "google.com" }

      expect(router.forbids?(URI("http://google.com"))).to be true
      expect(router.allows?(URI("http://google.com"))).to be false
    end
  end

  describe "#route" do
    context "with matching route" do
      it "returns the registered payload and parameters" do
        router.host "example.com", path: "/{barqux}", to: :foo

        uri = URI("http://example.com/42")
        method, params = router.route(uri)

        expect(method).to be :foo
        expect(params).to eq("barqux" => "42")
      end
    end

    context "with mismatching route" do
      it "returns false" do
        router.host "example.com", path: "/{barqux}", to: :foo

        uri = URI("http://google.com")

        expect(router.route(uri)).to be false
      end
    end

    context "with forbidden URI" do
      it "returns false" do
        router.host "example.com", path: "/{barqux}", to: :foo
        router.forbid.host "example.com"

        uri = URI("http://example.com")

        expect(router.route(uri)).to be false
      end
    end

    context "with nested conflicting target actions" do
      it "returns the expected parameters" do
        router.host "example.com", to: :foo do
          path "/{barqux}", to: :bar
        end

        uri = URI("http://example.com/42")
        method, params = router.route(uri)

        expect(method).to be :bar
        expect(params).to eq("barqux" => "42")
      end
    end
  end
end
